#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <assert.h>

#include "jpeglib.h"
#define PNG_DEBUG 3
#include "png.h"

static unsigned baseaddr[192];
static unsigned char grbuf[65536];

struct pixel {
	unsigned char r, g, b;
};

struct image {
	unsigned int h, w;
	struct pixel * p;
};

typedef struct pixel Pixel;
typedef struct pixel * PixelRef;
typedef struct image * ImageRef;

// static ImageRef testimage;

static void initBaseAddrs (void)
{
	unsigned int i, group_of_eight, line_of_eight, group_of_sixtyfour;
	
	for (i = 0; i < 192; ++i)
	{
		line_of_eight = i % 8;
		group_of_eight = (i % 64) / 8;
		group_of_sixtyfour = i / 64;
		
		baseaddr[i] = line_of_eight * 1024 + group_of_eight * 128 + group_of_sixtyfour * 40;
	}
}

static Pixel imageGetPixel (ImageRef i, unsigned int x, unsigned int y)
{
	Pixel z = {0,0,0};
	if (x >= i->w) return z;
	if (y >= i->h) return z;
	return i->p[(i->w * y) + x];
}

static void imagePutPixel (ImageRef i, unsigned int x, unsigned int y, Pixel p)
{
	if (x >= i->w) return;
	if (y >= i->h) return;
	i->p[(i->w * y) + x] = p;
}

static PixelRef imageGetPixelRef (ImageRef i, unsigned x, unsigned y)
{
	if (y >= i->h) return 0;
	if (x == i->w && y == 1+i->h) return 0;
	return i->p + (i->w * y) + x;
}

static void imageClear (ImageRef i)
{
	memset(i->p, 0, i->w * i->h * sizeof i->p[0]);
}

static int pixelDist (Pixel p1, Pixel p2)
{
	int d, pd;
	d = (int)p1.r - (int)p2.r;
	pd = d * d;
	d = (int)p1.g - (int)p2.g;
	pd = pd + d * d;
	d = (int)p1.b - (int)p2.b;
	return pd + d * d;
}

static int pixelDist7 (PixelRef pv1, PixelRef pv2)
{
	int i, d = 0;
	for (i = 0; i < 7; ++i)
	d = d + pixelDist(pv1[i], pv2[i]);
	return d;
}

static Pixel alterPixel (Pixel p, int dr, int dg, int db)
{
	dr += (int)p.r;
	dg += (int)p.g;
	db += (int)p.b;
	if (dr < 0) dr = 0; else if (dr > 255) dr = 255;
	if (dg < 0) dg = 0; else if (dg > 255) dg = 255;
	if (db < 0) db = 0; else if (db > 255) db = 255;
	p.r = dr;
	p.g = dg;
	p.b = db;
	return p;
}

struct obuf {
	unsigned char *p;
	unsigned char *n;
	unsigned int len, size;
};

static int initobuf (struct obuf *ob)
{
	ob->len = 0;
	ob->size = 4096;
	ob->n = ob->p = realloc(0, ob->size);
	return ob->p != 0;
}

static int ckobuf (struct obuf *ob, int len)
{
	unsigned char *p;
	if (ob->len + len <= ob->size) return 1;
	p = realloc(ob->p, ob->size + 4096);
	if (!p) return 0;
	ob->n = p + (ob->n - ob->p);
	ob->p = p;
	ob->size = ob->size + 4096;
	return 1;
}

static struct obuf * packBytes (void *vp, unsigned int len)
{
	unsigned int ilen, tlen, rlen, x;
	unsigned char *bp, *ip, *tp, *rp, b;
	static struct obuf ob;
	
	if (!initobuf(&ob)) return 0;
	
	bp = vp;
	ip = vp;
	ilen = len;
	
	while (ilen)
	{
		tp = ip;
		tlen = ilen;
		b = *tp++;
		while (--tlen && b == *tp) tp++;
		
		rlen = tp - ip;
		if (rlen > 2)
		{
			tlen = ip - bp;
			while (tlen > 0)
			{
				x = tlen;
				if (x >= 64) x = 64;
				tlen -= x;
				if (!ckobuf(&ob, x + 1)) return 0;
				ob.len += (x + 1);
				*(ob.n)++ = (unsigned char)(x - 1);
				while (x--)
				*(ob.n)++ = *bp++;
			}
			
			assert(bp == ip);
			
			if (rlen < 8 && rlen % 4)
			{
				if (!ckobuf(&ob, 2)) return 0;
				*(ob.n)++ = 0x40 | ((unsigned char)(rlen - 1));
				*(ob.n)++ = b;
				ob.len += 2;
				ilen -= rlen;
				ip += rlen;
			}
			else
			{
				rlen /= 4;
				if (rlen > 64) rlen = 64;
				if (!ckobuf(&ob, 2)) return 0;
				*(ob.n)++ = 0xC0 | ((unsigned char)(rlen - 1));
				*(ob.n)++ = b;
				ob.len += 2;
				ilen -= (rlen * 4);
				ip += (rlen * 4);
			}
			bp = ip;
			continue;
		}
		
		if (ilen >= 8)
		{
			rp = ip;
			tp = ip + 4;
			tlen = ilen - 4;
			while (tlen && *tp == *rp)
			{
				tp += 1;
				rp += 1;
				tlen -= 1;
			}
			
			rlen = tp - ip;
			if (rlen >= 8)
			{
				tlen = ip - bp;
				while (tlen > 0)
				{
					x = tlen;
					if (x >= 64) x = 64;
					tlen -= x;
					if (!ckobuf(&ob, x + 1)) return 0;
					ob.len += (x + 1);
					*(ob.n)++ = (unsigned char)(x - 1);
					while (x--)
					*(ob.n)++ = *bp++;
				}
				
				assert(bp == ip);
				
				rlen /= 4;
				if (rlen > 64) rlen = 64;
				if (!ckobuf(&ob, 5)) return 0;
				*(ob.n)++ = 0x80 | ((unsigned char)(rlen - 1));
				*(ob.n)++ = ip[0];
				*(ob.n)++ = ip[1];
				*(ob.n)++ = ip[2];
				*(ob.n)++ = ip[3];
				ob.len += 5;
				ilen -= (rlen * 4);
				ip += (rlen * 4);
				bp = ip;
				continue;
			}
		}
		
		ip += 1;
		ilen -= 1;
	}
	
	tlen = ip - bp;
	while (tlen > 0)
	{
		x = tlen;
		if (x >= 64) x = 64;
		tlen -= x;
		if (!ckobuf(&ob, x + 1)) return 0;
		ob.len += (x + 1);
		*(ob.n)++ = (unsigned char)(x - 1);
		while (x--)
		*(ob.n)++ = *bp++;
	}
	
	return &ob;
}

/*
static unsigned int unpackBytes (unsigned char *pd, unsigned int plen, void *vd, unsigned int ulen)
{
unsigned int len;
unsigned char *p = pd, *d = vd, *ud = vd;
	
while (plen > 0)
	{
	switch (*p & 0xC0) {
	case 0xC0:
	if (plen < 2) return d - ud;
	len = (unsigned int)((*p & 0x3F) + 1) * 4;
	goto case_0x40;
	case 0x80:
	if (plen < 5) return d - ud;
	len = (unsigned int)((*p & 0x3F) + 1);
	if (4 * len > ulen) return d - ud;
	plen -= 5;
	p += 1;
	while (len--) {
	d[0] = p[0];
	d[1] = p[1];
	d[2] = p[2];
	d[3] = p[3];
	d += 4;
	}
	p += 4;
	break;
	case 0x40:
	if (plen < 2) return d - ud;
	len = (unsigned int)((*p & 0x3F) + 1);
	case_0x40:
	if (len > ulen) return d - ud;
	plen -= 2;
	p += 1;
	while (len--)
	*d++ = *p;
	p += 1;
	break;
	case 0x00:
	len = (unsigned int)(*p + 1);
	if (plen < len + 1) return d - ud;
	if (len > ulen) return d - ud;
	plen = plen - len - 1;
	p += 1;
	while (len--)
	*d++ = *p++;
	break;
	default:
	return d - ud;
	}
	}
	
return d - ud;
}
*/

enum DitherType { EDIFF, ATKIN };

static void dither7 (enum DitherType alg, PixelRef buf, PixelRef pal)
{
	int z,i,d,bd,bi,dr,dg,db;
	
	for (z = 0; z < 7; z += 1) {
		bd = bi = 0x7fffffff; // big number
		for (i = 0; i < 4; ++i) {
			d = pixelDist(buf[z], pal[i]);
			if (d < bd) {
				bd = d;
				bi = i;
			}
		}
		
		dr = (int)buf[z].r - (int)pal[bi].r;
		dg = (int)buf[z].g - (int)pal[bi].g;
		db = (int)buf[z].b - (int)pal[bi].b;
		
		if (alg == ATKIN)
		{
			buf[z+1] = alterPixel(buf[z+1], dr/8, dg/8, db/8);
			buf[z+2] = alterPixel(buf[z+2], dr/8, dg/8, db/8);
		}
		else
		{
			buf[z+1] = alterPixel(buf[z+1], dr*7/16, dg*7/16, db*7/16);
		}
		
		buf[z] = pal[bi];
	}
}

static void ditherFromBuf (enum DitherType alg, PixelRef buf, ImageRef im, unsigned x, unsigned y)
{
	int i,j,k,dr,dg,db;
	Pixel ip;
	
	for (i = 0; i < 7; ++i)
	{
		ip = imageGetPixel(im, x+i, y);
		dr = (int)ip.r - (int)buf[i].r;
		dg = (int)ip.g - (int)buf[i].g;
		db = (int)ip.b - (int)buf[i].b;
		imagePutPixel(im, x+i, y, buf[i]);
		
		if (i == 6)
		{
			if (alg == ATKIN)
			{
				imagePutPixel(im, x+7, y, alterPixel(imageGetPixel(im, x+7, y), dr/8, dg/8, db/8));
				imagePutPixel(im, x+8, y, alterPixel(imageGetPixel(im, x+8, y), dr/8, dg/8, db/8));
			}
			else
			{
				imagePutPixel(im, x+7, y, alterPixel(imageGetPixel(im, x+7, y), dr*7/16, dg*7/16, db*7/16));
			}
		}
		
		if (y < im->h)
		{
			if (alg == ATKIN)
			{
				imagePutPixel(im, x+i, y+2, alterPixel(imageGetPixel(im, x+i, y+2), dr/8, dg/8, db/8));
				for (j = 0; j < 3; j += 1)
				{
					k = x + i + j - 1;
					imagePutPixel(im, k, y+1, alterPixel(imageGetPixel(im, k, y+1), dr/8, dg/8, db/8));
				}
			}
			else
			{
				static int f[3] = {3,5,1};
				for (j = 0; j < 3; j += 1)
				{
					k = x + i + j - 1;
					imagePutPixel(im, k, y+1, alterPixel(imageGetPixel(im, k, y+1), dr*f[j]/16, dg*f[j]/16, db*f[j]/16));
				}
			}
		}
	}
}

struct paletteEntry {
	Pixel p;
	int cnt;
	unsigned int smr, smg, smb;
};

struct paletteEntry grpal[] = {
	{{0x00,0x00,0x00},0,0,0,0},
	{{0xad,0x18,0x28},0,0,0,0},
	{{0x55,0x1b,0xe1},0,0,0,0},
	{{0xe8,0x2c,0xf8},0,0,0,0},
	{{0x01,0x73,0x63},0,0,0,0},
	{{0x7e,0x82,0x7f},0,0,0,0},
	{{0x34,0x85,0xfc},0,0,0,0},
	{{0xd1,0x95,0xff},0,0,0,0},
	{{0x33,0x6f,0x00},0,0,0,0},
	{{0xd0,0x81,0x01},0,0,0,0},
	{{0x7f,0x7e,0x77},0,0,0,0},
	{{0xfe,0x93,0xa3},0,0,0,0},
	{{0x1d,0xd6,0x09},0,0,0,0},
	{{0xae,0xea,0x22},0,0,0,0},
	{{0x5b,0xeb,0xd9},0,0,0,0},
	{{0xff,0xff,0xff},0,0,0,0}
};

static int comparePaletteOnCount (const void *a, const void *b)
{
	const struct paletteEntry *pa = a;
	const struct paletteEntry *pb = b;
	return pa->cnt - pb->cnt;
}

static int comparePixel (const void *a, const void *b)
{
	const Pixel *pa = a;
	const Pixel *pb = b;
	if (pa->r != pb->r) return (int)(pa->r) - (int)(pb->r);
	if (pa->g != pb->g) return (int)(pa->g) - (int)(pb->g);
	if (pa->b != pb->b) return (int)(pa->b) - (int)(pb->b);
	return 0;
}

static int comparePixelLumi (const void *a, const void *b)
{
	const Pixel *pa = a;
	const Pixel *pb = b;
	int ay, by;
	
	ay = 77 * (int)pa->r + 151 * (int)pa->g + 28 * (int)pa->b;
	by = 77 * (int)pb->r + 151 * (int)pb->g + 28 * (int)pb->b;
	
	return ay - by;
}

static unsigned char gsColor (unsigned char c)
{
	/*
	c &= 0xF0;
	c |= (c >> 4);
	return c;
*/
	/**/
	if (c > 0xF6) return 0xFF;
	if (c > 0xE5) return 0xEE;
	if (c > 0xD4) return 0xDD;
	if (c > 0xC3) return 0xCC;
	if (c > 0xB2) return 0xBB;
	if (c > 0xA1) return 0xAA;
	if (c > 0x90) return 0x99;
	if (c > 0x7F) return 0x88;
	if (c > 0x6E) return 0x77;
	if (c > 0x5D) return 0x66;
	if (c > 0x4C) return 0x55;
	if (c > 0x3B) return 0x44;
	if (c > 0x2A) return 0x33;
	if (c > 0x19) return 0x22;
	if (c > 0x08) return 0x11;
	return 0;
	/**/
}

static Pixel gsPixel (Pixel p)
{
	int r,g,b;
	int pr,pg,pb;
	
	pr = p.r;
	pg = p.g;
	pb = p.b;
	
	g = r = gsColor(p.r);
	if (pr >= pg) {
		while (pr - pg >= 9) {
			pr -= 17;
			g -= 17;
		}
		if (g < 0) g = 0;
	}
	else {
		while (pg - pr >= 9) {
			pr += 17;
			g += 17;
		}
		if (g > 0xFF) g = 0xFF;
	}

	b = g;
	if (pg >= pb) {
		while (pg - pb >= 9) {
			pg -= 17;
			b -= 17;
		}
		if (b < 0) b = 0;
	}
	else {
		while (pb - pg >= 9) {
			pg += 17;
			b += 17;
		}
		if (b > 0xFF) b = 0xFF;
	}
	
	p.r = (unsigned char)r;
	p.g = (unsigned char)g;
	p.b = (unsigned char)b;
	
	return p;
}

struct colorSpace {
	unsigned int npixels;
	unsigned char min[3];
	unsigned char max[3];
	unsigned char del[3];
	unsigned char pvec[320][3];
	int ridx;
	int rval;
};

/*
static int compareCSPixelRed (const void *a, const void *b) {
const unsigned char *pa = a, *pb = b;
return (int)(pa[0]) - (int)(pb[0]);
}

static int compareCSPixelBlu (const void *a, const void *b) {
const unsigned char *pa = a, *pb = b;
return (int)(pa[1]) - (int)(pb[1]);
}

static int compareCSPixelGrn (const void *a, const void *b) {
const unsigned char *pa = a, *pb = b;
return (int)(pa[2]) - (int)(pb[2]);
}

static int (* compareCSPixel[3])(const void *, const void *) = {
compareCSPixelRed, compareCSPixelBlu, compareCSPixelGrn
};

static int compareCSRange (const void *a, const void *b) {
const struct colorSpace *pa = a, *pb = b;
return pb->rval - pa->rval;
}

static void csRange (struct colorSpace *cs)
{
int x;
	
cs->min[0] = cs->min[1] = cs->min[2] = 0xff;
cs->max[0] = cs->max[1] = cs->max[2] = 0;
	
for (x = 0; x < cs->npixels; x += 1)
	{
	if (cs->pvec[x][0] < cs->min[0]) cs->min[0] = cs->pvec[x][0];
	if (cs->pvec[x][1] < cs->min[1]) cs->min[1] = cs->pvec[x][1];
	if (cs->pvec[x][2] < cs->min[2]) cs->min[2] = cs->pvec[x][2];
		
	if (cs->pvec[x][0] > cs->max[0]) cs->max[0] = cs->pvec[x][0];
	if (cs->pvec[x][1] > cs->max[1]) cs->max[1] = cs->pvec[x][1];
	if (cs->pvec[x][2] > cs->max[2]) cs->max[2] = cs->pvec[x][2];
	}
	
cs->del[0] = cs->max[0] - cs->min[0];
cs->del[1] = cs->max[1] - cs->min[1];
cs->del[2] = cs->max[2] - cs->min[2];
	
x = 0;
if (cs->del[1] > cs->del[0]) x = 1;
if (cs->del[2] > cs->del[x]) x = 2;
	
cs->ridx = x;
cs->rval = cs->del[x];
}

static void csSplit (struct colorSpace *old, struct colorSpace *new)
{
int i,x,y,z;
	
qsort(old->pvec, old->npixels, 3, compareCSPixel[old->ridx]);
	
x = y = z = old->npixels / 2;
while (y < old->npixels && old->pvec[y][old->ridx] == old->pvec[x][old->ridx]) y += 1;
while (z >= 0 && old->pvec[z][old->ridx] == old->pvec[x][old->ridx]) z -= 1;
if (z < x) z += 1;
if (y - x > x - z) x = z; else x = y;
	
for (i = x; i < old->npixels; i += 1)
	{
	new->pvec[i-x][0] = old->pvec[i][0];
	new->pvec[i-x][1] = old->pvec[i][1];
	new->pvec[i-x][2] = old->pvec[i][2];
	}
	
new->npixels = old->npixels - x;
old->npixels = x;
	
old->ridx = 0;
old->rval = 0;
new->ridx = 0;
new->rval = 0;
}

static void csAverage (struct colorSpace *cs)
{
unsigned int i,r,g,b;
	
r = g = b = 0;
for (i = 0; i < cs->npixels; i += 1) {
	r += (unsigned int)(cs->pvec[i][0]);
	g += (unsigned int)(cs->pvec[i][1]);
	b += (unsigned int)(cs->pvec[i][2]);
}
	
cs->del[0] = r / cs->npixels;
cs->del[1] = g / cs->npixels;
cs->del[2] = b / cs->npixels;
	
//    g = 0xffffffff;
//    for (i = 0; i < cs->npixels; i += 1) {
//    r = abs(cs->pvec[i][0] - cs->del[0]);
//    r = r + abs(cs->pvec[i][1] - cs->del[1]);
//    r = r + abs(cs->pvec[i][2] - cs->del[2]);
//    if (r < g) {
//    b = i;
//    g = r;
//    }
//    }
//	
//    cs->del[0] = cs->pvec[b][0];
//    cs->del[1] = cs->pvec[b][1];
//    cs->del[2] = cs->pvec[b][2];

}

static int GSPalette16 (Pixel pal[], Pixel line[])
{
int x,n;
int done, last;
static struct colorSpace cs[16];
	
n = 1;
cs[0].npixels = 320;
for (x = 0; x < 320; x += 1)
	{
	cs[0].pvec[x][0] = gsColor(line[x].r);
	cs[0].pvec[x][1] = gsColor(line[x].g);
	cs[0].pvec[x][2] = gsColor(line[x].b);
	}
	
done = 0;
while (n < 16 && !done)
	{
	for (x = 0; x < n; x += 1) {
	csRange(cs + x);
	}
		
	qsort(cs, n, sizeof cs[0], compareCSRange);
		
	last = n;
	for (x = 0; x < last && n < 16 && cs[x].rval; x += 1) {
	csSplit(cs + x, cs + n);
	n += 1;
	}
		
	done = (n == last);
	}
	
for (x = 0; x < n; x += 1) {
	csAverage(cs + x);
	pal[x].r = gsColor(cs[x].del[0]);
	pal[x].g = gsColor(cs[x].del[1]);
	pal[x].b = gsColor(cs[x].del[2]);
}
	
while (x < 16) {
	pal[x].r = pal[x].g = pal[x].b = 0;
	x += 1;
}
	
return n;
}

struct octreeNode {
unsigned int leaf;
unsigned int count;
unsigned int r,g,b;
unsigned int parent;
unsigned int child[8];
};

static unsigned int octreeMax = 0;
static struct octreeNode octree[8*320];

static unsigned int OctreeNewNode (void)
{
unsigned int i,n;
n = octreeMax;
octreeMax += 1;
octree[n].leaf = 0;
octree[n].count = 0;
octree[n].parent = 0;
for (i = 0; i < 8; i += 1)
	octree[n].child[i] = 0;
return n;
}

static void OctreeInsert (Pixel p)
{
unsigned int x,i,c,n;
unsigned char r,g,b;
	
i = 0;
octree[i].count += 1;
r = p.r; g = p.g; b = p.b;
	
for (x = 0; x < 8; x += 1)
	{
	octree[i].count += 1;
	c = ((r & 0x80) >> 5) | ((g & 0x80) >> 6) | ((b & 0x80) >> 7);
	r <<= 1; g <<= 1; b <<= 1;
	n = octree[i].child[c];
	if (0 == n) n = octree[i].child[c] = OctreeNewNode();
	octree[n].parent = i;
	i = n;
	}
	
octree[i].leaf = 1;
octree[i].count += 1;
octree[i].r = p.r;
octree[i].g = p.g;
octree[i].b = p.b;
}

static int OctreePrune (int maxleaves)
{
int level;
unsigned int idx, ci, ch, cidx[8];
unsigned int leafcount, mincount, minidx;
	
leafcount = 0;
mincount = 0x0fffffff;
	
idx = 0;
level = 0;
cidx[0] = 0;
	
while (level >= 0)
	{
	if (cidx[level] < 8) {
	ci = cidx[level];
	cidx[level] += 1;
	if (octree[idx].child[ci]) {
	level += 1;
	cidx[level] = 0;
	ch = octree[idx].child[ci];
	if (octree[ch].leaf) {
		leafcount += 1;
		if (octree[idx].count < mincount) {
		mincount = octree[idx].count;
		minidx = idx;
		}
	}
	idx = ch;
	}
	}
	else {
	idx = octree[idx].parent;
	level -= 1;
	}
	}
	
if (leafcount <= maxleaves) return leafcount;
	
octree[minidx].r = 0;
octree[minidx].g = 0;
octree[minidx].b = 0;
	
for (ci = 0; ci < 8; ci += 1)
	{
	if (octree[minidx].child[ci])
	{
	leafcount -= 1;
	idx = octree[minidx].child[ci];
	octree[minidx].child[ci] = 0;
	octree[minidx].r += (octree[idx].count * octree[idx].r);
	octree[minidx].g += (octree[idx].count * octree[idx].g);
	octree[minidx].b += (octree[idx].count * octree[idx].b);
	octree[idx].parent = 0;
	octree[idx].count = 0;
	octree[idx].leaf = 0;
	}
	}
	
octree[minidx].r /= octree[minidx].count;
octree[minidx].g /= octree[minidx].count;
octree[minidx].b /= octree[minidx].count;
octree[minidx].leaf = 1;
leafcount += 1;
	
return leafcount;
}

static int OctreePalette16 (Pixel pal[], Pixel line[])
{
int i,c;
	
octreeMax = 0;
OctreeNewNode();
for (i = 0; i < 320; i += 1) OctreeInsert(line[i]);
do
	c = OctreePrune(16);
while (c > 16);
for (c = i = 0; i < octreeMax; i += 1) {
	if (octree[i].leaf) {
	pal[c].r = octree[i].r;
	pal[c].g = octree[i].g;
	pal[c].b = octree[i].b;
	c += 1;
	}
}
return c;
}

static void GS320DitherLineMedianCut (enum DitherType alg, ImageRef im, int y)
{
int x,i,d,bd,bi,dr,dg,db,nc;
unsigned int tdr,tdg,tdb,ter;
Pixel line[322];
Pixel bpal[16];
	
memcpy(line, imageGetPixelRef(im, 0, y), 320 * sizeof line[0]);
	
tdr = tdg = tdb = 0;
	
nc = GSPalette16 (bpal, line);
//nc = OctreePalette16 (bpal, line);
		
for (x = 0; x < 320; x += 1)
	{
	bd = bi = 0x7fffffff; // big number
		
	for (i = 0; i < 16; i += 1) {
	d = pixelDist(line[x], bpal[i]);
	if (d < bd) {
	bd = d;
	bi = i;
	}
	}
		
	dr = (int)line[x].r - (int)bpal[bi].r;
	dg = (int)line[x].g - (int)bpal[bi].g;
	db = (int)line[x].b - (int)bpal[bi].b;
		
	tdr += abs(dr);
	tdg += abs(dg);
	tdb += abs(db);
		
	imagePutPixel(im, x, y, bpal[bi]);
		
	imagePutPixel(im, x+1, y,   alterPixel(imageGetPixel(im, x+1, y),   dr*7/16, dg*7/16, db*7/16));
	imagePutPixel(im, x-1, y+1, alterPixel(imageGetPixel(im, x-1, y+1), dr*3/16, dg*3/16, db*3/16));
	imagePutPixel(im, x,   y+1, alterPixel(imageGetPixel(im, x,   y+1), dr*5/16, dg*5/16, db*5/16));
	imagePutPixel(im, x+1, y+1, alterPixel(imageGetPixel(im, x+1, y+1), dr/16,   dg/16,   db/16));
	}
	
ter = tdr*tdr + tdg*tdg + tdb*tdb;
printf(" best error:%g  num. colors: %d\n", sqrt(ter), nc);
}

static Pixel mergePixels (Pixel p1, Pixel p2)
{
Pixel p;
p.r = ((int)p1.r + (int)p2.r) / 2;
p.g = ((int)p1.g + (int)p2.g) / 2;
p.b = ((int)p1.b + (int)p2.b) / 2;
return p;
}
*/

static int paletteLumiDistance (Pixel *p1, Pixel *p2)
{
	int i, d;
	
	for (d = i = 0; i < 16; i += 1)
	d += abs(comparePixelLumi(p1+i, p2+i));

	return d;
}

static void GS320DitherPalette (Pixel *bpal, enum DitherType alg, ImageRef im, int y, int quality)
{
	int i,j,x,d,bd,bi,dr,dg,db;
	unsigned int tdr,tdg,tdb,ter,ber,ler,bco;
	int resets;
	Pixel line[322];
	struct paletteEntry pal[16];
	
	if (quality < 1) quality = 1;
	
	resets = 0;
	srandom(time(0));
	
	for (j = 0; j < 16; j += 1) {
		pal[j].p = imageGetPixel(im, j*20+10, y);
	}
	
	ter = ber = 0xffffffff; // big number
	memset(bpal, 0, 16 * sizeof bpal[0]);
	
	for (j = 0; j < 32768; j += 1)
	{
		tdr = tdg = tdb = 0;
		
		memcpy(line, imageGetPixelRef(im, 0, y), 320 * sizeof line[0]);
		
		for (i = 0; i < 16; i += 1)
		{
			pal[i].p = gsPixel(pal[i].p);
			/*
		pal[i].p.r = gsColor(pal[i].p.r);
		pal[i].p.g = gsColor(pal[i].p.g);
		pal[i].p.b = gsColor(pal[i].p.b);
	*/
			pal[i].cnt = 0;
			pal[i].smr = 0;
			pal[i].smg = 0;
			pal[i].smb = 0;
		}
		
		for (x = 0; x < 320; x += 1)
		{
			bd = bi = 0x7fffffff; // big number
			
			for (i = 0; i < 16; i += 1) {
				d = pixelDist(line[x], pal[i].p);
				if (d < bd) {
					bd = d;
					bi = i;
				}
			}
			
			pal[bi].cnt += 1;
			pal[bi].smr += (int)(line[x].r);
			pal[bi].smg += (int)(line[x].g);
			pal[bi].smb += (int)(line[x].b);
			
			dr = (int)line[x].r - (int)pal[bi].p.r;
			dg = (int)line[x].g - (int)pal[bi].p.g;
			db = (int)line[x].b - (int)pal[bi].p.b;
			
			tdr += abs(dr);
			tdg += abs(dg);
			tdb += abs(db);
			
			if (alg == ATKIN) {
				line[x+1] = alterPixel(line[x+1], dr/8, dg/8, db/8);
				line[x+2] = alterPixel(line[x+2], dr/8, dg/8, db/8);
			}
			else {
				line[x+1] = alterPixel(line[x+1], dr*7/16, dg*7/16, db*7/16);
			}
		}
		
		ler = ter;
		ter = tdr*tdr + tdg*tdg + tdb*tdb;
		//ter = tdr*tdr + tdg*tdg/4 + tdb*tdb/9;
		
		if (0 == ter) {
			// bail out on "perfect" palette
			ber = 0;
			bco = 1;
			break;
		}
		
		if (ter < ber) {
			bco = 0;
			for (x = 0; x < 16; ++x) {
				bpal[x] = pal[x].p;
				if (pal[x].cnt) bco += 1;
			}
			ber = ter;
			resets = 0;
		}
		
		if (ter >= ler) {
			//printf("total error = %8g (reset)\n", sqrt(ter));
			if (++resets == quality) j = 0x0fffffff;
			ter = 0xffffffff;
			for (i = 0; i < 16; i += 1) {
				pal[i].p = bpal[i];
				pal[i].p.r = pal[i].p.r + ((random() % 20) - 10);
				pal[i].p.g = pal[i].p.g + ((random() % 20) - 10);
				pal[i].p.b = pal[i].p.b + ((random() % 20) - 10);
			}
		}
		else
		{
			qsort(pal, 16, sizeof pal[0], comparePaletteOnCount);
			
			for (i = 15; i >= 0 && pal[i].cnt; i -= 1)
			{
				pal[i].p.r = pal[i].smr / pal[i].cnt;
				pal[i].p.g = pal[i].smg / pal[i].cnt;
				pal[i].p.b = pal[i].smb / pal[i].cnt;
			}
			
			for (d = 15 - i, x = i - d; i >= 0 && i > x; i -= 1)
			{
				pal[i].p.r = ((int)(pal[i].p.r) + (int)(pal[i+d].p.r)) / 2;
				pal[i].p.g = ((int)(pal[i].p.g) + (int)(pal[i+d].p.g)) / 2;
				pal[i].p.b = ((int)(pal[i].p.b) + (int)(pal[i+d].p.b)) / 2;
			}
			
			if (i >= 0)
			{
				for (d = 320 / (i + 1); i >= 0; i -= 1)
				{
					pal[i].p = imageGetPixel(im, d*i+d/2, y);
				}
			}
			
			/*
		for (x = 0; x < (15 - i); x += 1) {
		if (pal[x].cnt) break;
		pal[x].p = pal[15-x].p;
		}
			
		while (x <= i) {
		pal[x].p.r = ((int)(pal[x].p.r) + (int)(pal[15].p.r)) / 2;
		pal[x].p.g = ((int)(pal[x].p.g) + (int)(pal[15].p.g)) / 2;
		pal[x].p.b = ((int)(pal[x].p.b) + (int)(pal[15].p.b)) / 2;
		x += 1;
		}
			
		while (x < 16) {
		pal[x].p.r = pal[x].smr / pal[x].cnt;
		pal[x].p.g = pal[x].smg / pal[x].cnt;
		pal[x].p.b = pal[x].smb / pal[x].cnt;
		x += 1;
		}
	*/
		}
		
		//printf("%2d) total error = %u %u %u = %8g\n", j, tdr, tdg, tdb, sqrt(ter))
	}
	
	//printf("%2d) least error:%g  num. colors: %u\n", y, sqrt(ber), bco);
}

static Pixel greyPals[15][16];
static unsigned char lumiVals[15][16] = {
	{0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF},
	{0x22, 0x22, 0x22, 0x22, 0x66, 0x66, 0x66, 0x66, 0xAA, 0xAA, 0xAA, 0xAA, 0xEE, 0xEE, 0xEE, 0xEE},
	{0x00, 0x00, 0x00, 0x22, 0x22, 0x22, 0x44, 0x66, 0x99, 0xBB, 0xDD, 0xDD, 0xDD, 0xFF, 0xFF, 0xFF},
	{0x00, 0x00, 0x11, 0x11, 0x22, 0x22, 0x33, 0x33, 0x44, 0x44, 0x55, 0x55, 0x66, 0x66, 0x77, 0x77},
	{0x11, 0x11, 0x22, 0x22, 0x33, 0x33, 0x44, 0x44, 0x55, 0x55, 0x66, 0x66, 0x77, 0x77, 0x88, 0x88},
	{0x22, 0x22, 0x33, 0x33, 0x44, 0x44, 0x55, 0x55, 0x66, 0x66, 0x77, 0x77, 0x88, 0x88, 0x99, 0x99},
	{0x33, 0x33, 0x44, 0x44, 0x55, 0x55, 0x66, 0x66, 0x77, 0x77, 0x88, 0x88, 0x99, 0x99, 0xAA, 0xAA},
	{0x44, 0x44, 0x55, 0x55, 0x66, 0x66, 0x77, 0x77, 0x88, 0x88, 0x99, 0x99, 0xAA, 0xAA, 0xBB, 0xBB},
	{0x55, 0x55, 0x66, 0x66, 0x77, 0x77, 0x88, 0x88, 0x99, 0x99, 0xAA, 0xAA, 0xBB, 0xBB, 0xCC, 0xCC},
	{0x66, 0x66, 0x77, 0x77, 0x88, 0x88, 0x99, 0x99, 0xAA, 0xAA, 0xBB, 0xBB, 0xCC, 0xCC, 0xDD, 0xDD},
	{0x77, 0x77, 0x88, 0x88, 0x99, 0x99, 0xAA, 0xAA, 0xBB, 0xBB, 0xCC, 0xCC, 0xDD, 0xDD, 0xEE, 0xEE},
	{0x88, 0x88, 0x99, 0x99, 0xAA, 0xAA, 0xBB, 0xBB, 0xCC, 0xCC, 0xDD, 0xDD, 0xEE, 0xEE, 0xFF, 0xFF},
	{0x00, 0x22, 0x44, 0x66, 0x88, 0xAA, 0xBB, 0xBB, 0xCC, 0xCC, 0xDD, 0xDD, 0xEE, 0xEE, 0xFF, 0xFF},
	{0x00, 0x00, 0x11, 0x11, 0x22, 0x22, 0x33, 0x33, 0x44, 0x44, 0x66, 0x77, 0x99, 0xBB, 0xDD, 0xFF},
	{0x00, 0x22, 0x44, 0x55, 0x66, 0x66, 0x77, 0x77, 0x88, 0x88, 0x99, 0x99, 0xAA, 0xBB, 0xDD, 0xFF}
};

static void initGreyPals (void)
{
	int i,j;
	
	for (i = 0; i < 15; i += 1) {
		for (j = 0; j < 16; j += 1) {
			greyPals[i][j].r = lumiVals[i][j];
			greyPals[i][j].g = lumiVals[i][j];
			greyPals[i][j].b = lumiVals[i][j];
		}
	}
}

enum fileType { FTC00002, FTC10001, FTC10002 };

static void GS320Dither (enum DitherType alg, ImageRef im, int quality, enum fileType ft, FILE *f)
{
	int d,i,x,y,bi,bd,dr,dg,db;
	Pixel bpal[200][16];
	Pixel p;
	unsigned char b, line[160], palnum[200];
	struct obuf packline[200];
	
	
	//y = 70;
	for (y = 0; y < 200; y += 1)
	{
		palnum[y] = 0;
		printf("\rquantizing with quality %d: line %3d", quality, y+1);
		fflush(stdout);
		
		GS320DitherPalette(&(bpal[y][0]), alg, im, y, quality);
		//GS320DitherLineMedianCut(alg, im, y);
		
		qsort(&(bpal[y][0]), 16, sizeof bpal[0][0], comparePixelLumi);
	}
	putchar('\n');
	
	for (y = 0; y < 200; y += 1)
	{
		b = 0;
		for (x = 0; x < 320; x += 1)
		{
			p = imageGetPixel(im, x, y);
			bd = bi = 0x7fffffff; // big number
			
			for (i = 0; i < 16; i += 1)
			{
				d = pixelDist(p, bpal[y][i]);
				if (d < bd) 
				{
					bd = d;
					bi = i;
				}
			}
			
			dr = (int)p.r - (int)bpal[y][bi].r;
			dg = (int)p.g - (int)bpal[y][bi].g;
			db = (int)p.b - (int)bpal[y][bi].b;
			
			imagePutPixel(im, x, y, bpal[y][bi]);
			
			if (ft == FTC00002 || ft == FTC10002)
			{
				b = 16 * b | (unsigned char)bi;
				if (x & 1) {
					line[x/2] = b;
					b = 0;
				}
			}
			
			if (alg == ATKIN)
			{
				imagePutPixel(im, x+1, y,   alterPixel(imageGetPixel(im, x+1, y  ), dr/8, dg/8, db/8));
				imagePutPixel(im, x+2, y,   alterPixel(imageGetPixel(im, x+2, y  ), dr/8, dg/8, db/8));
				imagePutPixel(im, x-1, y+1, alterPixel(imageGetPixel(im, x-1, y+1), dr/8, dg/8, db/8));
				imagePutPixel(im, x,   y+1, alterPixel(imageGetPixel(im, x,   y+1), dr/8, dg/8, db/8));
				imagePutPixel(im, x+1, y+1, alterPixel(imageGetPixel(im, x+1, y+1), dr/8, dg/8, db/8));
				imagePutPixel(im, x,   y+2, alterPixel(imageGetPixel(im, x,   y+2), dr/8, dg/8, db/8));
			}
			else
			{
				imagePutPixel(im, x+1, y,   alterPixel(imageGetPixel(im, x+1, y),   dr*7/16, dg*7/16, db*7/16));
				imagePutPixel(im, x-1, y+1, alterPixel(imageGetPixel(im, x-1, y+1), dr*3/16, dg*3/16, db*3/16));
				imagePutPixel(im, x,   y+1, alterPixel(imageGetPixel(im, x,   y+1), dr*5/16, dg*5/16, db*5/16));
				imagePutPixel(im, x+1, y+1, alterPixel(imageGetPixel(im, x+1, y+1), dr/16,   dg/16,   db/16));
			}
		}
		
		if (ft == FTC10002)
		{
			fwrite(line, 1, 160, f);
		}
		else if (ft == FTC00002)
		{
			struct obuf *op;
			op = packBytes(line, 160);
			if (!op) {
				fputs("Out of memory in GS320Dither.\n", stderr);
				for (x = 0; x < y; x += 1)
				free(packline[x].p);
				return;
			}
			packline[y] = *op;
		}
	}
	
	for (y = 0; y < 200; y += 1) {
		bd = 0x0fffffff; // big number
		for (i = 0; i < 15; i += 1) {
			d = paletteLumiDistance(&(bpal[y][0]), &(greyPals[i][0]));
			if (d < bd) {
				bd = d;
				bi = i;
			}
		}
		palnum[y] = bi;
	}
	
	if (ft == FTC00002)
	{
		int blocklen;
		char buf[6];
		char idmain[] = {'\004','M','A','I','N'};
		char idmultipal[] = {'\010','M','U','L','T','I','P','A','L'};
		
		// MAIN block
		
		blocklen = 1329;
		for (y = 0; y < 200; y += 1)
		blocklen += packline[y].len;
		
		//printf("main block len : %d\n", blocklen);
		
		buf[0] = blocklen & 0xFF;
		buf[1] = (blocklen / 256) & 0xFF;
		buf[2] = (blocklen / 256 / 256) & 0xFF;
		buf[3] = (blocklen / 256 / 256 / 256) & 0xFF;
		fwrite(buf, 1, 4, f);
		
		fwrite(idmain, 1, 5, f);
		
		buf[0] = 0x00;  // modeword from
		buf[1] = 0x00;  //   Master SCB (?)
		buf[2] = 0x40;  // pixels per line
		buf[3] = 0x01;  //   == 320 (0x0140)
		buf[4] = 0x10;  // palettes
		buf[5] = 0x00;  //   == 16 (0x0010)
		fwrite(buf, 1, 6, f);
		
		for (y = 0; y < 16; y += 1) {
			for (x = 0; x <= 16; x += 1) {
				line[2*x] = (greyPals[y][x].g & 0xF0) | (greyPals[y][x].b & 0x0F);
				line[2*x+1] = greyPals[y][x].r & 0x0F;
			}
			fwrite(line, 1, 32, f);
		}
		
		buf[0] = 0xC8;  // num scanlines
		buf[1] = 0x00;  //  == 200 (0x00C8)
		fwrite(buf, 1, 2, f);
		
		for (y = 0; y < 200; y += 1) {
			buf[0] = packline[y].len;
			buf[1] = 0x00;
			buf[2] = palnum[y];
			buf[3] = 0x00;
			fwrite(buf, 1, 4, f);
		}
		
		for (y = 0; y < 200; y += 1) {
			fwrite(packline[y].p, 1, packline[y].len, f);
		}
#if 1
		// MULTIPAL block
		
		buf[0] = 0x0F;
		buf[1] = 0x19;
		buf[2] = 0x00;
		buf[3] = 0x00;
		fwrite(buf, 1, 4, f);
		fwrite(idmultipal, 1, 9, f);
		
		buf[0] = 0xC8;  // num palettes
		buf[1] = 0x00;  //  == 200 (0x00C8)
		fwrite(buf, 1, 2, f);
		
		for (y = 0; y < 200; y += 1) {
			for (x = 0; x <= 16; x += 1) {
				line[2*x] = (bpal[y][x].g & 0xF0) | (bpal[y][x].b & 0x0F);
				line[2*x+1] = bpal[y][x].r & 0x0F;
			}
			fwrite(line, 1, 32, f);
		}
#endif
	}
	else if (ft == FTC10002)
	{
		for (y = 0; y < 200; y += 1) {
			for (x = 0; x <= 16; x += 1) {
				line[2*x] = (bpal[y][15-x].g & 0xF0) | (bpal[y][15-x].b & 0x0F);
				line[2*x+1] = bpal[y][15-x].r & 0x0F;
			}
			fwrite(line, 1, 32, f);
		}
	}
}

static unsigned int dhgrDither (enum DitherType alg, ImageRef im, unsigned char *dhgr, int pack)
{
	static Pixel pal[] = {
		{   1,   4,   8}, // 0
		{  32,  54, 212}, // 1
		{  51, 111,   0}, // 2
		{   7, 168, 225}, // 3
		{  99,  77,   0}, // 4
		{ 126, 126, 126}, // 5
		{  67, 200,   0}, // 6
		{  93, 248, 133}, // 7
		{ 148,  12, 125}, // 8
		{ 188,  55, 255}, // 9
		{ 126, 126, 126}, // A
		{ 158, 172, 255}, // B
		{ 249,  86,  29}, // C
		{ 255, 129, 236}, // D
		{ 221, 206,  23}, // E
		{ 248, 250, 244}  // F
	};
	
	int y,x,d,i,bd,bi,dr,dg,db;
	unsigned int ad;
	Pixel p;
	struct obuf *ob;
	
	for (y = 0; y < im->h; y += 1)
	{
		for (x = 0; x < im->w; x += 1)
		{
			bd = bi = 0x7fffffff; // big number
			p = imageGetPixel(im, x, y);
			
			for (i = 0; i < 16; ++i) {
				d = pixelDist(p, pal[i]);
				if (d < bd) {
					bd = d;
					bi = i;
				}
			}
			
			dr = (int)p.r - (int)pal[bi].r;
			dg = (int)p.g - (int)pal[bi].g;
			db = (int)p.b - (int)pal[bi].b;
			
			imagePutPixel(im, x, y, pal[bi]);

			if (alg == ATKIN)
			{
				imagePutPixel(im, x+1, y,   alterPixel(imageGetPixel(im, x+1, y  ), dr/8, dg/8, db/8));
				imagePutPixel(im, x+2, y,   alterPixel(imageGetPixel(im, x+2, y  ), dr/8, dg/8, db/8));
				imagePutPixel(im, x-1, y+1, alterPixel(imageGetPixel(im, x-1, y+1), dr/8, dg/8, db/8));
				imagePutPixel(im, x,   y+1, alterPixel(imageGetPixel(im, x,   y+1), dr/8, dg/8, db/8));
				imagePutPixel(im, x+1, y+1, alterPixel(imageGetPixel(im, x+1, y+1), dr/8, dg/8, db/8));
				imagePutPixel(im, x,   y+2, alterPixel(imageGetPixel(im, x,   y+2), dr/8, dg/8, db/8));
			}
			else
			{
				imagePutPixel(im, x+1, y,   alterPixel(imageGetPixel(im, x+1, y),   dr*7/16, dg*7/16, db*7/16));
				imagePutPixel(im, x-1, y+1, alterPixel(imageGetPixel(im, x-1, y+1), dr*3/16, dg*3/16, db*3/16));
				imagePutPixel(im, x,   y+1, alterPixel(imageGetPixel(im, x,   y+1), dr*5/16, dg*5/16, db*5/16));
				imagePutPixel(im, x+1, y+1, alterPixel(imageGetPixel(im, x+1, y+1), dr/16,   dg/16,   db/16));
			}
		}
	}

	for (y = 0; y < im->h; y += 1) {
		ad = baseaddr[y];
		for (x = 0; x < im->w; x += 7) {
			d = 0;
			for (i = 0; i < 7; i += 1) {
				p = imageGetPixel(im, x + i, y);
				switch (p.r) {
					/**/
				case   1: d = (d >> 4) | (0x00 << 24); break;
				case  32: d = (d >> 4) | (0x01 << 24); break;
				case  51: d = (d >> 4) | (0x02 << 24); break;
				case   7: d = (d >> 4) | (0x03 << 24); break;
				case  99: d = (d >> 4) | (0x04 << 24); break;

				case  67: d = (d >> 4) | (0x06 << 24); break;
				case  93: d = (d >> 4) | (0x07 << 24); break;
				case 148: d = (d >> 4) | (0x08 << 24); break;
				case 188: d = (d >> 4) | (0x09 << 24); break;

				case 158: d = (d >> 4) | (0x0b << 24); break;
				case 249: d = (d >> 4) | (0x0c << 24); break;
				case 255: d = (d >> 4) | (0x0d << 24); break;
				case 221: d = (d >> 4) | (0x0e << 24); break;
				case 248: d = (d >> 4) | (0x0f << 24); break;

				case 126: d = (d >> 4) | ((d & (1 << 23)) ? (0x0a << 24) : (0x05 << 24)); break;
					/**/
				default: if (p.r) printf("Unknown pixel (%2.2x)\n", p.r); d = d >> 4; break;
				}
			}
			
			for (i = 0; i < 4; ++i) {
				unsigned char bits;
				unsigned int  addr;
				addr = ad + i/2;
				bits = (unsigned char)(d & 0x7f);
				if (i & 1) {
					dhgr[0x2000+addr] = bits;
				}
				else {
					dhgr[addr] = bits;
				}
				d = d >> 7;
			}
			
			ad += 2;
		}
	}
	
	if (!pack) return 0x4000;
	
	ob = packBytes(dhgr, 0x4000);
	if (!ob) return 0;
	
	memcpy(dhgr, ob->p, ob->len);
	free(ob->p);
	return ob->len;
}

static void hgrBytes (Pixel *p, unsigned char *hgr)
{
	int x, y, s, i;
	unsigned char b, h;
	
	i = s = 0;
	for (y = 0; y < 2; ++y)
	{
		b = h = 0;
		for (x = 0; x < 7; ++x)
		{
			if (s == 0)
			{
				s = 1;
				switch (p[i].r)
				{
				case 0x34:	/* blue */
					h = 0x80;
					/* continue */
				case 0xe8:	/* violet */
				case 0xff:	/* white */
					b = (b >> 1) | 0x80;
					break;
				case 0x1d:	/* green */
				case 0xd0:	/* orange */
				default:	/* black */
					b = b >> 1;
					break;
				}
			}
			else /* s == 1 */
			{
				s = 0;
				i = i + 1;
				switch (p[i].r)
				{
				case 0xd0:	/* orange */
					h = 0x80;
					/* continue */
				case 0x1d:	/* green */
				case 0xff:	/* white */
					b = (b >> 1) | 0x80;
					break;
				case 0xe8:	/* violet */
				case 0x34:	/* blue */
				default:	/* black */
					b = b >> 1;
					break;
				}
			}
		}
		
		hgr[y] = (b >> 1) | h;
	}
}

static unsigned int hgrDither (enum DitherType alg, ImageRef src, unsigned char *hgr, int pack)
{
	static Pixel pal1[] = {{0,0,0},{0xe8,0x2c,0xf8},{0x1d,0xd6,0x09},{0xff,0xff,0xff}};
	static Pixel pal2[] = {{0,0,0},{0x34,0x85,0xfc},{0xd0,0x81,0x01},{0xff,0xff,0xff}};
	int y,x,z,d1,d2;
	unsigned int ad, i;
	PixelRef pp;
	Pixel buf1[9], buf2[9];
	struct obuf *ob;
	
	for (y = 0; y < src->h; y += 1) {
		for (x = 0; x < src->w; x += 7) {
			for (z = 0; z < 9; z += 1) {
				buf1[z] = imageGetPixel(src, x+z, y);
				buf2[z] = imageGetPixel(src, x+z, y);
			}
			dither7(alg, buf1, pal1);
			dither7(alg, buf2, pal2);
			pp = imageGetPixelRef(src, x, y);
			d1 = pixelDist7(pp, buf1);
			d2 = pixelDist7(pp, buf2);
			if (d1 < d2)
			ditherFromBuf(alg, buf1, src, x, y);
			else
			ditherFromBuf(alg, buf2, src, x, y);
		}
	}
	
	for (y = 0; y < src->h; y += 1) {
		i = 0;
		ad = baseaddr[y];
		for (x = 0; x < src->w; x += 7) {
			hgrBytes(imageGetPixelRef(src,x,y), hgr+ad+i);
			i += 2;
		}
	}
	
	if (!pack) return 0x2000;

	ob = packBytes(hgr, 0x2000);
	if (!ob) return 0;
	
	memcpy(hgr, ob->p, ob->len);
	free(ob->p);
	return ob->len;
}

static ImageRef imageNew (unsigned int w, unsigned int h)
{
	ImageRef ip;
	ip = malloc(sizeof *ip);
	if (!ip) return 0;
	ip->w = w;
	ip->h = h;
	ip->p = malloc(ip->h * ip->w * sizeof ip->p[0]);
	if (ip->p) return ip;
	free(ip);
	return 0;
}

static ImageRef imageCopy (ImageRef im)
{
	ImageRef cp;
	cp = imageNew(im->w, im->h);
	if (cp)	memcpy(cp->p, im->p, im->h * im->w * sizeof im->p[0]);
	return cp;
}

static void imageFree (ImageRef im)
{
	free(im->p);
	free(im);
}

static void imageScale (ImageRef src, ImageRef dst, float asprat)
{
	int sw, sh, bx, sx, sy;
	float srcw, srch, dstw, dsth, r, g, b;
	float xfactor, yfactor, scalefactor, x, y, nx, ny, ix, iy, nix, niy, area;
	Pixel p;
	
	imageClear(dst);
	
	srcw = (float)src->w;
	srch = (float)src->h;
	dstw = (float)dst->w;
	dsth = (float)dst->h;
	
	if (srch > dsth || srcw > dstw)
	{
		xfactor = srcw / dstw / asprat;
		yfactor = srch / dsth;
		
		scalefactor = xfactor;
		if (yfactor > scalefactor) scalefactor = yfactor;
		
		sw = (int)(srcw / scalefactor / asprat);
		sh = (int)(srch / scalefactor);
		sx = (int)(((float)dst->w - sw) / 2);
		sy = (int)(((float)dst->h - sh) / 2);
		bx = sx;
		
		xfactor = scalefactor * asprat;
		
		printf("xfact:%g yfact:%g\n", xfactor, yfactor);
		printf("srcw:%g srch:%g sw:%d sh:%d\n", srcw, srch, sw, sh);
		printf("dstw:%g dsth:%g sx:%d sy:%d\n", dstw, dsth, sx, sy);
		
		for (y = 0.0; y < srch; y = ny)
		{
			ny = y + scalefactor;
			for (x = 0.0; x < srcw; x = nx)
			{
				nx = x + xfactor;
				r = g = b = 0;
				
				if (sy == 52 && sx == 161) {
					putchar('.');
				}
				
				for (iy = y; iy < ny; iy = niy)
				{
					niy = (int)(iy + 1.0);
					
					for (ix = x; ix < nx; ix = nix)
					{
						nix = (int)(ix + 1.0);
						area = 1.0;
						
						if ((iy - (int)iy) > 0.00001) {
							area = niy - iy;
						}
						else if (iy + 1.0 >= ny) {
							area = ny - iy;
						}
						
						if ((ix - (int)ix) > 0.00001) {
							area = area * (nix - ix);
						}
						else if (ix + 1.0 >= nx) {
							area = area * (nx - ix);
						}
						
						p = imageGetPixel(src, (int)ix, (int)iy);
						
						r += (area * (float)p.r);
						g += (area * (float)p.g);
						b += (area * (float)p.b);
					}
				}
				
				r = r / xfactor / scalefactor;
				g = g / xfactor / scalefactor;
				b = b / xfactor / scalefactor;
				
				if (r < 0.0) r = 0.0; else if (r >= 256.0) r = 255.0;
				if (g < 0.0) g = 0.0; else if (g >= 256.0) g = 255.0;
				if (b < 0.0) b = 0.0; else if (b >= 256.0) b = 255.0;
				
				p.r = (int)(r+0.1);
				p.g = (int)(g+0.1);
				p.b = (int)(b+0.1);
				
				imagePutPixel(dst, sx, sy, p);
				sx += 1;
			}
			sx = bx;
			sy += 1;
		}
	}
}

struct cmap {
	char cx[8];
	Pixel px;
};

static int cmapcmp (const void *a, const void *b)
{
	char *x, *y;
	x = ((struct cmap *)a)->cx;
	y = ((struct cmap *)b)->cx;
	return strcmp(x,y);
}

static char * xpmPixel(char *xc, int xd, int ps, int idx)
{
	static char cb[10];
	char *p;
	
	p = cb;
	while (idx) {
		*p++ = xc[idx % xd];
		idx /= xd;
		ps -= 1;
	}
	
	while (ps) {
		*p++ = ' ';
		ps -= 1;
	}
	
	*p = '\0';
	return cb;
}

static int hgrToXPM (char *fname, ImageRef im, int exp)
{
	int x,y,i,nc,ps,xd;
	FILE *f;
	char *c;
	Pixel p;
	PixelRef cur, nxt;
	ImageRef cp;
	static char nb[1000];
	static char *xc =	" .+@#$%&*=-;>,')!~{]^/(_:<[}|1234567890"
	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ`";
	
	f = fopen(fname, "w");
	if (!f) return -1;
	cp = imageCopy(im);
	if (!cp) return -1;
	
	qsort(cp->p, cp->h * cp->w, sizeof cp->p[0], comparePixel);
	cur = nxt = cp->p;
	
	for (x = 0; x < (cp->h * cp->w); x += 1) {
		if (comparePixel(cur, nxt))	{
			cur += 1;
			cur[0] = nxt[0];
		}
		
		nxt += 1;
	}
	
	ps = 0;
	xd = strlen(xc);
	nc = cur - cp->p;
	
	while (nc) {
		ps += 1;
		nc /= xd;
	}
	
	if (0 == ps) ps = 1;
	nc = (cur - cp->p) + 1;
	
	strncpy(nb,fname,999);
	c = strrchr(nb, '.');
	if (c) *c = '_';
	
	fprintf(f, "/* XPM */\nstatic char * %s[] = {\n\"%d %d %d %d\",\n", nb, exp*im->w, 2*im->h, nc, ps);
	
	nxt = cp->p;
	do {
		fprintf(f, "\"%s\tc #%2.2X%2.2X%2.2X\",\n", xpmPixel(xc, xd, ps, nxt - cp->p), nxt->r, nxt->g, nxt->b);
		nxt += 1;
	} while (nxt <= cur);
	
	for (y = 0; y < im->h; ) {
		fprintf(f,"\"");
		for (x = 0; x < im->w; x += 1) {
			p = imageGetPixel(im, x, y);
			nxt = bsearch(&p, cp->p, nc, sizeof p, comparePixel);
			
			if (!nxt) {
				fprintf(f, "Error! Color (%2.2x,%2.2x,%2.2x) not found in palette!\n", p.r, p.g, p.b);
				imageFree(cp);
				fclose(f);
				return -1;
			}
			
			c = xpmPixel(xc, xd, ps, nxt - cp->p);
			
			for (i = exp; i; i -= 1)
			fputs(c,f);
		}
		
		fprintf(f,"\",\n\"");
		for (x = 0; x < im->w; x += 1) {
			p = imageGetPixel(im, x, y);
			nxt = bsearch(&p, cp->p, nc, sizeof p, comparePixel);
			c = xpmPixel(xc, xd, ps, nxt - cp->p);
			for (i = exp; i; i -= 1)
			fputs(c,f);
		}
		
		y += 1;
		if (y < im->h) fprintf(f,"\",\n");
		else fprintf(f,"\"};");
	}
	
	imageFree(cp);
	fclose(f);
	return 0;
}


void write_png(char *fname, ImageRef im, int exp)
{
	int x,y,i;
	Pixel p;
	png_structp png_ptr;
	png_infop info_ptr;
	png_bytep row;
	
	/* create file */
	FILE *fp = fopen(fname, "wb");
	if (!fp)
	printf("[write_png_file] File %s could not be opened for writing", fname);


	/* initialize stuff */
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

	if (!png_ptr)
	printf("[write_png_file] png_create_write_struct failed");

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
	printf("[write_png_file] png_create_info_struct failed");

	if (setjmp(png_jmpbuf(png_ptr)))
	printf("[write_png_file] Error during init_io");

	png_init_io(png_ptr, fp);


	/* write header */
	if (setjmp(png_jmpbuf(png_ptr)))
	printf("[write_png_file] Error during writing header");

	png_set_IHDR(png_ptr, info_ptr, exp*im->w, 2*im->h,
	8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
	PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	png_write_info(png_ptr, info_ptr);

	
	for (y=0; y < im->h; y++){
		row = (png_bytep) malloc(3 * exp *im->w * sizeof(png_byte));
		for (x = 0; x < im->w; x++) {
			png_byte* ptr = &(row[x*3*exp]);
			p = imageGetPixel(im, x, y);
			for (i = 0; i < exp; i++){
				ptr[0+i*3] = p.r;
				ptr[1+i*3] = p.g;
				ptr[2+i*3] = p.b;
			}
		}
		png_write_row(png_ptr, row);
		png_write_row(png_ptr, row);
	}

	/* end write */
	if (setjmp(png_jmpbuf(png_ptr)))
	printf("[write_png_file] Error during end of write");

	png_write_end(png_ptr, NULL);

	/* cleanup heap allocation */
	free(row);

	fclose(fp);
}

void write_bmp(char *fname, ImageRef im)
{
	int x,y,i,j;
	Pixel p;
	FILE *f;
	unsigned char *img = NULL;
	int filesize = 54 + 3 * im->w * im->h;
	img = (unsigned char *)malloc(3 * im->w * im->h);

	for(i=0; i<im->w; i++)
	{
		for(j=0; j<im->h; j++)
		{
			x=i; y=(im->h-1)-j;
			p = imageGetPixel(im, x, y);
			img[(x+y*im->w)*3+2] = p.r;
			img[(x+y*im->w)*3+1] = p.g;
			img[(x+y*im->w)*3+0] = p.b;
		}
	}

	unsigned char bmpfileheader[14] = {'B','M', 0,0,0,0, 0,0, 0,0, 54,0,0,0};
	unsigned char bmpinfoheader[40] = {40,0,0,0, 0,0,0,0, 0,0,0,0, 1,0, 24,0};
	unsigned char bmppad[3] = {0,0,0};

	bmpfileheader[ 2] = (unsigned char)(filesize    );
	bmpfileheader[ 3] = (unsigned char)(filesize>> 8);
	bmpfileheader[ 4] = (unsigned char)(filesize>>16);
	bmpfileheader[ 5] = (unsigned char)(filesize>>24);

	bmpinfoheader[ 4] = (unsigned char)(im->w    );
	bmpinfoheader[ 5] = (unsigned char)(im->w>> 8);
	bmpinfoheader[ 6] = (unsigned char)(im->w>>16);
	bmpinfoheader[ 7] = (unsigned char)(im->w>>24);
	
	bmpinfoheader[ 8] = (unsigned char)(im->h    );
	bmpinfoheader[ 9] = (unsigned char)(im->h>> 8);
	bmpinfoheader[10] = (unsigned char)(im->h>>16);
	bmpinfoheader[11] = (unsigned char)(im->h>>24);

	f = fopen(fname,"wb");
	fwrite(bmpfileheader,1,14,f);
	fwrite(bmpinfoheader,1,40,f);
	for(i=0; i<im->h; i++)
	{
		fwrite(img+(im->w*(im->h-i-1)*3),3,im->w,f);
		fwrite(bmppad,1,(4-(im->w*3)%4)%4,f);
	}
	fclose(f);
}

void write_bmp16(char *fname, ImageRef im)
{
	static Pixel pal[] = {
		{   1,   4,   8}, // 0
		{ 148,  12, 125}, // 1
		{  99,  77,   0}, // 2
		{ 249,  86,  29}, // 3
		{  51, 111,   0}, // 4
		{ 126, 126, 126}, // 5
		{  67, 200,   0}, // 6
		{ 221, 206,  23}, // 7
		{  32,  54, 212}, // 8
		{ 188,  55, 255}, // 9
		{ 126, 126, 126}, // A
		{ 255, 129, 236}, // B
		{   7, 168, 225}, // C
		{ 158, 172, 255}, // D
		{  93, 248, 133}, // E
		{ 248, 250, 244}  // F
	};
	
	int x,y,i,j,k;
	Pixel p;
	FILE *f;
	unsigned char *img = NULL;
	int filesize = 54 +  (im->w * im->h)/2;
	img = (unsigned char *)malloc((im->w * im->h)/2);
	memset(img,0,(im->w * im->h)/2);
	unsigned char palette[16][4];

	for(i=0; i<im->w; i++)
	{
		for(j=0; j<im->h; j++)
		{
			x=i/2; y=(im->h-1)-j;
			p = imageGetPixel(im, i, y);
			for(k=0; k<16; k++){
				if(pixelDist(p, pal[k]) == 0)
					break;
			}
			if(i%2)
				img[(x+y*im->w/2)] |= k;
			else
				img[(x+y*im->w/2)] |= k<<4;
		}
	}

	unsigned char bmpfileheader[14] = {'B','M', 0,0,0,0, 0,0, 0,0, 118,0,0,0};
	unsigned char bmpinfoheader[40] = {40,0,0,0, 0,0,0,0, 0,0,0,0, 1,0, 4,0};
	unsigned char bmppad[3] = {0,0,0};

	bmpfileheader[ 2] = (unsigned char)(filesize    );
	bmpfileheader[ 3] = (unsigned char)(filesize>> 8);
	bmpfileheader[ 4] = (unsigned char)(filesize>>16);
	bmpfileheader[ 5] = (unsigned char)(filesize>>24);

	bmpinfoheader[ 4] = (unsigned char)(im->w    );
	bmpinfoheader[ 5] = (unsigned char)(im->w>> 8);
	bmpinfoheader[ 6] = (unsigned char)(im->w>>16);
	bmpinfoheader[ 7] = (unsigned char)(im->w>>24);
	
	bmpinfoheader[ 8] = (unsigned char)(im->h    );
	bmpinfoheader[ 9] = (unsigned char)(im->h>> 8);
	bmpinfoheader[10] = (unsigned char)(im->h>>16);
	bmpinfoheader[11] = (unsigned char)(im->h>>24);

	f = fopen(fname,"wb");
	fwrite(bmpfileheader,1,14,f);
	fwrite(bmpinfoheader,1,40,f);
	for(i=0; i< 16; i++)
	{
		palette[i][0]=pal[i].b;
		palette[i][1]=pal[i].g;
		palette[i][2]=pal[i].r;
	}
	fwrite(palette,4,16,f);
	for(i=0; i<im->h; i++)
	{
		fwrite(img+(im->w*(im->h-i-1)/2),1,im->w/2,f);
		fwrite(bmppad,1,(4-(im->w/2)%4)%4,f);
	}
	fclose(f);
}

static ImageRef imageFromJPG (FILE *f)
{
	ImageRef ip = 0;
	PixelRef pp = 0;
	JSAMPLE * scanline = 0;
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, f);

	jpeg_read_header(&cinfo, TRUE);
	jpeg_start_decompress(&cinfo);

	ip = imageNew(cinfo.output_width, cinfo.output_height);
	if (!ip) goto emalloc;
	// fprintf(stderr, "w:%d h:%d\n", ip->w, ip->h);

	scanline = malloc(ip->w * cinfo.out_color_components * sizeof(JSAMPLE));
	if (!scanline) goto emalloc;

	pp = ip->p;
	// fprintf(stderr, "first: %p   limit: %p\n", ip->p, ip->p + ip->h * ip->w);
	while (cinfo.output_scanline < cinfo.output_height)
	{
		int icnt;
		JSAMPLE *sp = scanline;

		jpeg_read_scanlines(&cinfo, &scanline, 1);

		// fprintf(stderr, "[%d:pp=%p] ", cinfo.output_scanline, pp); fflush(stderr);

		if (cinfo.out_color_components == 1)
		{
			for (icnt = ip->w; icnt; --icnt, ++pp)
			{
				pp->r = pp->g = pp->b = *sp++;
			}
		}
		else
		{
			for (icnt = ip->w; icnt; --icnt, ++pp)
			{
				pp->r = *sp++;
				pp->g = *sp++;
				pp->b = *sp++;
			}
		}
	}

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

	return ip;

emalloc:
	if (scanline) free(scanline);
	if (ip->p) free(ip->p);
	if (ip) free(ip);
	return 0;
}

static ImageRef imageFromPNG(FILE *fp)
{
	ImageRef ip = 0;
	PixelRef pp = 0;
	png_structp png_ptr;
	png_infop info_ptr;
	png_bytep scanline;
	png_byte color_type;
	int y, i;
	char header[8];    // 8 is the maximum size that can be checked

	/* open file and test for it being a png */
	fread(header, 1, 8, fp);
	if (png_sig_cmp((png_const_bytep)header, 0, 8))
	printf("[read_png_file] File is not recognized as a PNG file");


	/* initialize stuff */
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

	if (!png_ptr)
	printf("[read_png_file] png_create_read_struct failed");

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
	printf("[read_png_file] png_create_info_struct failed");

	if (setjmp(png_jmpbuf(png_ptr)))
	printf("[read_png_file] Error during init_io");

	png_init_io(png_ptr, fp);
	png_set_sig_bytes(png_ptr, 8);

	png_read_info(png_ptr, info_ptr);
	color_type = png_get_color_type(png_ptr, info_ptr);
	ip = imageNew(png_get_image_width(png_ptr, info_ptr), png_get_image_height(png_ptr, info_ptr));
	if (!ip) goto emalloc;	
	png_read_update_info(png_ptr, info_ptr);
	
	scanline = malloc(png_get_rowbytes(png_ptr,info_ptr));
	if (!scanline) goto emalloc;
	/* read file */
	if (setjmp(png_jmpbuf(png_ptr)))
	printf("[read_png_file] Error during read_image");
	pp = ip->p;
	for (y=0; y<ip->h; y++){
		png_bytep sp = scanline;
		png_read_row(png_ptr, scanline, NULL);
		for (i = 0; i < ip->w; i++, ++pp)
		{
			pp->r = *sp++;
			pp->g = *sp++;
			pp->b = *sp++;
			if(color_type == PNG_COLOR_TYPE_RGB_ALPHA) sp++;
		}
	}
	return ip;

emalloc:
	if (scanline) free(scanline);
	if (ip->p) free(ip->p);
	if (ip) free(ip);
	return 0;		
}

static char sbuf[65536];
static ImageRef imageFromXPM (FILE *f)
{
	int c, n, ncol, pwid;
	unsigned int r, g, b;
	char *t;
	char pb[8];
	ImageRef ip = 0;
	PixelRef pp;
	Pixel zp = {0,0,0};
	struct cmap *cp = 0, *ep;
	
	c = 0;
	while (c != '"' && !feof(f)) c = fgetc(f);
	if (feof(f)) return 0;
	
	ip = malloc(sizeof *ip);
	if (!ip) goto emalloc;
	
	if (!fgets(sbuf, 65536, f)) goto expm;
	sscanf(sbuf, "%d %d %d %d", &(ip->w), &(ip->h), &ncol, &pwid);
	//printf("w:%d h:%d ncol:%d pwid:%d\n", ip->w, ip->h, ncol, pwid);
	
	ip->p = malloc(ip->h * ip->w * sizeof ip->p[0]);
	if (!ip->p) goto emalloc;
	cp = malloc(ncol * sizeof *cp);
	if (!cp) goto emalloc;
	
	for (n = 0; n < ncol; ++n) {
		if (!fgets(sbuf, 65536, f)) goto expm;
		cp[n].px = zp;
		memset(cp[n].cx, 0, 8);
		strncpy(cp[n].cx, sbuf+1, pwid);
		t = strchr(sbuf+1+pwid, 'c');
		if (t && *(t+2) == '#') {
			sscanf(t+3, "%2x%2x%2x", &r, &g, &b);
			cp[n].px.r = r;
			cp[n].px.g = g;
			cp[n].px.b = b;
		}
	}
	
	qsort(cp, ncol, sizeof *cp, cmapcmp);
	pp = ip->p;
	memset(pb, 0, 8);
	for (n = 0; n < ip->h; ++n) {
		if (!fgets(sbuf, 65536, f)) goto expm;
		t = sbuf+1;
		for (c = 0; c < ip->w; ++c) {
			strncpy(pb, t+pwid*c, pwid);
			ep = bsearch(pb, cp, ncol, sizeof *cp, cmapcmp);
			if (ep) *pp++ = ep->px;
			else *pp++ = zp;
		}
	}
	
	return ip;

expm:
	printf("Error reading XPM file\n");
emalloc:
	if (ip->p) free(ip->p);
	if (ip) free(ip);
	if (cp) free(cp);
	return 0;
}

static void print_usage() {
    printf("\nUsage: tohgr [options] a.xpm [options] b.png ... [options] z.jpg \n\n");
    printf("Options accepted:\n\n");
    printf("-hgr   : convert following files to hires images\n");
	printf("-dhgr  : convert following files to double hires images\n");
    printf("-pack  : produce packed HGR/DHGR files (Filetype $08)\n");
    printf("-3200  : convert following files to 320x200 super hires 3200 color images\n");
    printf("-q [integer]  : quality value for super hires conversion (default = 16)\n");
    printf("-atkin : use Atkinson dithering instead of Floyd-Steinberg dithering\n");
	printf("-xpm   : output preview files in XPM format instead of PNG\n");
	printf("-bmp   : output DHGR file in BMP format\n");	
}

enum GRType { HGR, DHGR, GR, DGR, GS640, GS320, GS800C, GS3200C };
static char nbuf[1024];

enum ArgState { BASE, QUAL };

int main (int argc, const char * argv[]) {
	int i,did;
	char *dot;
	FILE *inf;
	FILE *grf;
	ImageRef im, sm, gm;
	enum GRType grt;
	enum DitherType alg;
	enum ArgState ast;
	int quality, pack, xpm, bmp;
	unsigned int len;
	
	did = 0;
	grt = HGR;
	ast = BASE;
	alg = EDIFF;
	quality = 16;
	pack = 0;
	xpm = 0;
	bmp = 0;
	initGreyPals();
	initBaseAddrs();
	sm = imageNew(140, 192);
	gm = imageNew(320, 200);
	
	for (i = 1; i < argc; ++i)
	{
		if (ast == BASE && argv[i][0] == '-')
		{
			if (!strcmp(argv[i], "-hgr"))
			grt = HGR;
			else if (!strcmp(argv[i], "-dhgr"))
			grt = DHGR;
			else if (!strcmp(argv[i], "-3200"))
			grt = GS320;
			else if (!strcmp(argv[i], "-atkin"))
			alg = ATKIN;
			else if (!strcmp(argv[i], "-q"))
			ast = QUAL;
			else if (!strcmp(argv[i], "-pack"))
			pack = 1;
			else if (!strcmp(argv[i], "-xpm"))
			xpm = 1;
			else if (!strcmp(argv[i], "-bmp"))
			bmp = 1;			
		}
		else if (ast == QUAL)
		{
			int q;
			if (sscanf(argv[i], "%d", &q)) {
				quality = q;
			}
			ast = BASE;
		}
		else
		{
			strncpy(nbuf, argv[i], 1000);
			dot = strrchr(nbuf, '.');
			if (dot)
			{
				if (0 == strcmp(dot, ".xpm") || 0 == strcmp(dot, ".XPM"))
				{
					did = 1;
					inf = fopen(argv[i], "r");
					if (!inf) {
						printf("Failed to open file '%s'\n", argv[i]);
						continue;
					}
					im = imageFromXPM(inf);
					if (!im) {
						fclose(inf);
						printf("Can't read XPM file '%s'\n", argv[i]);
						continue;
					}
				}
				else if (0 == strcmp(dot, ".jpg") || 0 == strcmp(dot, ".JPG"))
				{
					did = 1;
					inf = fopen(argv[i], "rb");
					if (!inf) {
						printf("Failed to open file '%s'\n", argv[i]);
						continue;
					}
					im = imageFromJPG(inf);
					if (!im) {
						fclose(inf);
						printf("Can't read JPG file '%s'\n", argv[i]);
						continue;
					}
				}
				else if (0 == strcmp(dot, ".png") || 0 == strcmp(dot, ".PNG"))
				{
					did = 1;
					inf = fopen(argv[i], "rb");
					if (!inf) {
						printf("Failed to open file '%s'\n", argv[i]);
						continue;
					}
					im = imageFromPNG(inf);
					if (!im) {
						fclose(inf);
						printf("Can't read PNG file '%s'\n", argv[i]);
						continue;
					}
				}
				else 
				{
					printf("Format not supported");
					return 0;
				}
				
				switch (grt) {
				case HGR:
					imageScale(im, sm, 1.904762);
					len = hgrDither(alg, sm, grbuf, pack);
#if 1
					if(xpm){
						strcpy(dot, "_hgr.xpm");
						hgrToXPM(nbuf, sm, 4);
					} else{
						strcpy(dot, "_hgr.png");
						write_png(nbuf, sm, 4);
					}
#endif
					if (pack)
					strcpy(dot, ".08.4000");
					else
					strcpy(dot, ".hgr");
					grf = fopen(nbuf, "wb");
					fwrite(grbuf, 1, len, grf);
					fclose(grf);
					break;
					
				case DHGR:
					imageScale(im, sm, 1.904762);
					len = dhgrDither(alg, sm, grbuf, pack);
#if 1
					if(xpm){
						strcpy(dot, "_dhgr.xpm");
						hgrToXPM(nbuf, sm, 4);
					} else{
						strcpy(dot, "_dhgr.png");
						write_png(nbuf, sm, 4);
					}
#endif
					if(bmp){
						strcpy(dot, "_dhgr.bmp");
						write_bmp16(nbuf, sm);
					}
					else
					{
						if (pack)
							strcpy(dot, ".08.4001");
						else
							strcpy(dot, ".dhgr");
						grf = fopen(nbuf, "wb");
						fwrite(grbuf, 1, len, grf);
						fclose(grf);
					}
					break;
					
				case GS320:
					imageScale(im, gm, 5.0/6.0);
#if 1
					strcpy(dot, ".C0.0002");
					grf = fopen(nbuf, "wb");
#endif
					GS320Dither (alg, gm, quality, FTC00002, grf);
					fclose(grf);
					if(xpm){
						strcpy(dot, "_3200.xpm");
						hgrToXPM(nbuf, gm, 2);
					} else {
						strcpy(dot, "_3200.png");
						write_png(nbuf, gm, 2);
					}
					break;

				default:
					print_usage();
					break;
				}
				
				fclose(inf);

				if (im->p) free(im->p);
				if (im) free(im);
			}
		}
	}

	if (did == 0)
	print_usage();

	return 0;
}
